import importlib


class LazyLoader:
    """Lazy loader that defers the import of a module or class until it's actually used."""

    def __init__(self, module_name, class_name=None):
        self.module_name = module_name
        self.class_name = class_name
        self.module = None

    def __getattr__(self, name):
        if self.module is None:
            self.module = importlib.import_module(self.module_name)
        return getattr(self.module, name)

    def __class__(self):
        return self.module

    def __call__(self, *args, **kwargs):
        """Allow the loader to be called and delegate to the imported class."""
        cls = self._load()
        return cls(*args, **kwargs)

    def _load(self):
        if self.module is None:
            self.module = importlib.import_module(self.module_name)
        if self.class_name:
            return getattr(self.module, self.class_name)
        return self.module



import json

from gcp_pal.utils import log, ClientHandler, ModuleHandler, get_default_arg


class CloudScheduler:

    def __init__(self, name=None, project=None, location=None, service_account=None):
        self.service_account = service_account
        self.project = project or get_default_arg("project")
        self.location = location or get_default_arg("location")
        if isinstance(name, str) and name.startswith("projects/"):
            name = name.split("/")[-1]
        self.name = name
        self.parent = f"projects/{self.project}/locations/{self.location}"
        self.full_name = f"{self.parent}/jobs/{self.name}"

        self.scheduler = ModuleHandler("google.cloud").please_import(
            "scheduler_v1", who_is_calling="CloudScheduler"
        )
        self.types = self.scheduler.types
        self.client = ClientHandler(self.scheduler.CloudSchedulerClient).get()
        self.exceptions = ModuleHandler("google.api_core.exceptions").please_import(
            who_is_calling="CloudScheduler"
        )

    def __repr__(self):
        return f"CloudScheduler({self.name})"

    def ls(self, full_name=False):
        """
        List all jobs in the project.

        Args:
        - full_name (bool): Return full name or just the job name.

        Returns:
        - List[google.cloud.scheduler_v1.Job]: List of jobs.
        """
        jobs = self.client.list_jobs(parent=self.parent)
        if full_name:
            output = [x.name for x in jobs]
        else:
            output = [x.name.split("/")[-1] for x in jobs]
        return output

    def get(self, name=None):
        """
        Get a job by name.

        Args:
        - name (str): Name of the job.

        Returns:
        - google.cloud.scheduler_v1.Job: Job object.
        """
        name = name or self.full_name
        if not name.startswith("projects/"):
            name = f"{self.parent}/jobs/{name}"
        return self.client.get_job(name=name)

    def exists(self):
        """
        Check if a job exists.

        Returns:
        - bool: True if the job exists.
        """
        try:
            self.client.get_job(name=self.full_name)
            return True
        except Exception:
            return False

    def create(
        self,
        schedule,
        target,
        time_zone="UTC",
        payload={},
        description=None,
        service_account=None,
    ):
        """
        Create a job.

        Args:
        - schedule (str): Schedule in cron format.
        - target (str): Target of the job. Can be a URL (HTTP trigger) or a Pub/Sub topic.
        - payload (str): Payload to send to the job.
        - time_zone (str): Time zone. Default is "UTC".
        - description (str): Description of the job.
        - service_account (str): Service account email. If "DEFAULT", uses the default service account PROJECT@PROJECT.iam.gserviceaccount.com.

        Returns:
        - google.cloud.scheduler_v1.Job: Job object.
        """
        oauth_token = None
        oidc_token = None
        service_account = service_account or self.service_account
        if service_account == "DEFAULT":
            service_account = f"{self.project}@{self.project}.iam.gserviceaccount.com"
        if service_account:
            oauth_token = self.types.OAuthToken(service_account_email=service_account)
            oidc_token = self.types.OidcToken(service_account_email=service_account)
        job = self.types.Job(
            name=self.full_name,
            schedule=schedule,
            time_zone=time_zone,
            description=description,
        )
        payload = json.dumps(payload).encode("utf-8")
        if not target.startswith("http"):
            job.pubsub_target = self.types.PubsubTarget(topic_name=target, data=payload)
        else:
            http_method = self.types.HttpMethod.POST
            job.http_target = self.types.HttpTarget(
                uri=target,
                http_method=http_method,
                body=payload,
                oauth_token=oauth_token,
                oidc_token=oidc_token,
            )

        if self.exists():
            output = self.client.update_job(job=job)
            log(f"CloudScheduler - Job updated: {self.name}.")
        else:
            output = self.client.create_job(parent=self.parent, job=job)
            log(f"CloudScheduler - Job created: {self.name}.")
        return output

    def delete(self, errors="ignore"):
        """
        Delete a job.

        Args:
        - errors (str): If job does not exist, ignore or raise an error.

        Returns:
        - google.cloud.scheduler_v1.Job: Job object.
        """
        try:
            output = self.client.delete_job(name=self.full_name)
            log(f"CloudScheduler - Job deleted: {self.name}.")
            return output
        except Exception as e:
            if errors == "ignore":
                log(f"CloudScheduler - Job {self.name} does not exist to delete.")
                return None
            else:
                raise e

    def status(self):
        """
        Get the status of the job.

        Returns:
        - str: Job status.
        """
        got = self.get()
        if got.last_attempt_time is None:
            return "Has not run yet"
        status = got.status
        try:
            code = status.code
        except AttributeError:
            code = 0
        if code == 0:
            return "Success"
        elif code == 2:
            return "Failed"
        return status

    def state(self):
        """
        Get the state of the job.

        Returns:
        - str: Job state. Possible values are:
            - ENABLED: The job is enabled and can be executed.
            - PAUSED: The job is paused. It will not run.
            - UPDATE_FAILED: The job failed to update.
            - DISABLED: The job is disabled. It will not run.
        """
        got = self.get()
        state = got.state.name
        return state

    def run(self, force=True):
        """
        Run the job.

        Args:
        - force (bool): If job is paused, force run it by resuming it.

        Returns:
        - google.cloud.scheduler_v1.Job: Job object.
        """
        try:
            output = self.client.run_job(name=self.full_name)
        except self.exceptions.FailedPrecondition:
            if force:
                self.resume()
                output = self.client.run_job(name=self.full_name)
            else:
                raise
        log(f"CloudScheduler - Job ran: {self.name}.")
        return output

    def pause(self):
        """
        Pause the job.

        Returns:
        - google.cloud.scheduler_v1.Job: Job object.
        """
        output = self.client.pause_job(name=self.full_name)
        log(f"CloudScheduler - Job paused: {self.name}.")
        return output

    def resume(self):
        """
        Resume the job.

        Returns:
        - google.cloud.scheduler_v1.Job: Job object.
        """
        output = self.client.resume_job(name=self.full_name)
        log(f"CloudScheduler - Job resumed: {self.name}.")
        return output


if __name__ == "__main__":
    CloudScheduler("test-job").create(
        schedule="*/5 * * * *",
        time_zone="UTC",
        target="https://123ex123am123ple123.com",
        payload={"key": "value"},
    )
    print(CloudScheduler().ls())
    print(CloudScheduler("test-job").status())
    print(CloudScheduler("test-job").state())
    CloudScheduler("test-job").run()
    print(CloudScheduler("test-job").status())
    CloudScheduler("test-job").pause()
    print(CloudScheduler("test-job").state())
    CloudScheduler("test-job").delete()



import time
import datetime

from gcp_pal.utils import log, ClientHandler, ModuleHandler, get_default_arg


class LogEntry:
    def __init__(self, project, log_name, resource, severity, message, timestamp):
        self.project = project
        self.log_name = log_name
        self.resource = resource
        self.severity = severity
        self.message = message
        self.timestamp = timestamp
        self.time_zone = timestamp.tzinfo
        self.timestamp_str = (
            timestamp.isoformat(sep=" ", timespec="milliseconds").split("+")[0]
            + f" {self.time_zone}"
        )
        self.message_str = self._parse_message()

    def to_dict(self):
        return {
            "project": self.project,
            "log_name": self.log_name,
            "resource": self.resource,
            "severity": self.severity,
            "message": self.message,
            "timestamp": self.timestamp,
            "timestamp_str": self.timestamp_str,
            "message_str": self.message_str,
        }

    def to_api_repr(self):
        return {
            "project": self.project,
            "log_name": self.log_name,
            "resource": self.resource,
            "severity": self.severity,
            "message": self.message,
            "timestamp": self.timestamp,
            "timestamp_str": self.timestamp_str,
            "message_str": self.message_str,
        }

    def _parse_message(self):
        if isinstance(self.message, dict) and "message" in self.message:
            return self.message["message"]
        return self.message

    def __str__(self):
        return f"LogEntry - [{self.timestamp_str}] {self.message_str}"

    def __repr__(self):
        return f"LogEntry - [{self.timestamp_str}] {self.message_str}"


class Logging:

    def __init__(self, project=None):
        self.project = project or get_default_arg("project")
        self.filters = []

        self.loggingClient = ModuleHandler("google.cloud").please_import(
            "logging", who_is_calling="Logging"
        )
        self.client = ClientHandler(self.loggingClient).get(project=self.project)

    def ls(
        self,
        query=None,
        severity=None,
        limit=None,
        time_start=None,
        time_end=None,
        time_range=None,
        order_by="timestamp desc",
    ):
        """
        List all logs in a project

        Args:
        - query (str): Query for filtering results
        - severity (str): Severity level
        - limit (int): Number of logs to return
        - time_start (datetime.datetime): Start time for logs
        - time_end (datetime.datetime): End time for logs
        - time_range (int): Time range in hours. Can't be used with time_start and time_end.
        - order_by (str): Order logs by ascending or descending. Default is desc.

        Returns:
        - (list[dict]): List of log entries
        """
        if time_range:
            time_end = datetime.datetime.now(datetime.timezone.utc)
            time_start = time_end - datetime.timedelta(hours=time_range)
        query = self._generate_query(query, severity, time_start, time_end)
        log(f"Logging - Filter: {query}")
        if limit is None and query is None:
            limit = 100
        logs = self.client.list_entries(
            filter_=query, max_results=limit, order_by=order_by
        )
        output = []
        for log_entry in logs:
            output.append(
                LogEntry(
                    project=log_entry.resource.labels["project_id"],
                    log_name=log_entry.log_name,
                    resource=log_entry.resource,
                    severity=log_entry.severity,
                    message=log_entry.payload,
                    timestamp=log_entry.timestamp,
                )
            )
        return output

    def stream(self, query=None, severity=None, time_start=None, interval=5):
        """
        Stream logs in a project in real-time by polling the log entries.

        Args:
        - query (str): Query for filtering the logs. Default is None.
        - severity (str): Severity level. Default is None.
        - time_start (datetime.datetime): Start time for logs. Default is now.
        - interval (int): Polling interval in seconds. Default is 5 seconds.

        Yields:
        - (LogEntry): Yield log entries as they are found.
        """
        if time_start is None:
            time_start = datetime.datetime.now(datetime.timezone.utc)

        last_end_time = time_start
        time_zone = last_end_time.tzinfo
        time_start_str = last_end_time.isoformat(sep=" ").replace(
            "+00:00", f" {time_zone}"
        )
        log(f"Logging - Start Time: {time_start_str}. Streaming...")

        while True:
            current_time = datetime.datetime.now(datetime.timezone.utc)
            # Buffer to account for GCP log latency
            log_latency = datetime.timedelta(seconds=10)
            time_end = current_time - log_latency
            if time_end <= last_end_time:
                time.sleep(interval)  # Wait until window is positive
                continue

            log_filter = self._generate_query(
                query, severity, last_end_time.isoformat(), time_end.isoformat()
            )
            logs = self.client.list_entries(filter_=log_filter)

            for log_entry in logs:
                le = LogEntry(
                    project=log_entry.resource.labels["project_id"],
                    log_name=log_entry.log_name,
                    resource=log_entry.resource,
                    severity=log_entry.severity,
                    message=log_entry.payload,
                    timestamp=log_entry.timestamp,
                )
                print(le)

            last_end_time = time_end  # Shift the time window
            time.sleep(interval)

    def _generate_query(
        self, query=None, severity=None, time_start=None, time_end=None
    ):
        if query:
            self.filters.append(query)
        if severity:
            filter_str = f"severity={severity}"
            self.filters.append(filter_str)
        if time_start:
            if not isinstance(time_start, str):
                time_start = time_start.isoformat()
            filter_str = f'timestamp>="{time_start}"'
            self.filters.append(filter_str)
        if time_end:
            if not isinstance(time_end, str):
                time_end = time_end.isoformat()
            filter_str = f'timestamp<="{time_end}"'
            self.filters.append(filter_str)
        query = " AND ".join(self.filters)
        self.filters = []  # Reset filters
        query = None if query == "" else query
        return query


if __name__ == "__main__":
    Logging().stream()



import threading

from gcp_pal.utils import ModuleHandler, ClientHandler, log, get_default_arg


class ArtifactRegistry:

    _sentinel = object()

    def __init__(
        self,
        path: str = "",
        project: str = _sentinel,
        location: str = _sentinel,
        repository: str = None,
        image: str = None,
        version: str = None,
        tag: str = None,
    ):
        """
        Initialize an ArtifactRegistry object.

        Args:
        - path (str): The path to the resource. This follows the hierarchy `repository/image/version/tag`.
                      If a full path is given (starting with `'projects/'`), the `project`, `location`, `repository`,
                      `image`, and `tag` will be extracted from the path.
                      The path can be either one of the forms:
            - `projects/PROJECT/locations/LOCATION/repositories/REPOSITORY/packages/IMAGE/versions/sha256:VERSION`
            - `REPOSITORY/IMAGE/VERSION`
            - `REPOSITORY/IMAGE:TAG`
        - project (str): The GCP project ID. Defaults to the `PROJECT` environment variable or the default auth project.
        - location (str): The location of the Artifact Registry. Defaults to `'europe-west2'`.
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image (SHA256 hash).
        - tag (str): The tag of the image (e.g. `'latest'`). If a tag is provided, the `version` will be ignored.
        """
        if isinstance(path, str) and path.startswith("projects/"):
            path = path.split("/")
            path = "/".join(path[1::2])
            path = path.replace("%2F", "/")
            # Extract project and location from path and leave path to be repository/image/package
            try:
                project = path.split("/")[0]
            except IndexError:
                pass
            try:
                location = path.split("/")[1]
            except IndexError:
                # If path is provided as 'projects/project', assume location is intentionally left out
                location = None
            try:
                path = "/".join(path.split("/")[2:])
            except IndexError:
                pass

        self.project = (
            project if project is not self._sentinel else get_default_arg("project")
        )
        self.location = (
            location if location is not self._sentinel else get_default_arg("location")
        )
        self.repository = repository
        self.image = image
        self.tag = tag
        self.version = version
        try:
            self.repository = path.split("/")[0] if path != "" else self.repository
        except IndexError:
            pass
        try:
            self.image = path.split("/")[1:]
            if len(self.image) > 1:
                self.image = "/".join(self.image[:-1])
            elif self.image == []:
                self.image = None or image
            else:
                self.image = self.image[0]
        except IndexError:
            pass
        try:
            self.version = path.split("sha256:")[1]
        except IndexError:
            pass
        try:
            self.tag = self.image.split(":")[1]
            self.image = self.image.split(":")[0]
        except (IndexError, AttributeError):
            pass
        if self.tag:
            self.version = None
        self.level = self._get_level()
        self.path = self._get_path()
        self.artifactregistry_v1 = ModuleHandler(
            "google.cloud.artifactregistry_v1"
        ).please_import(who_is_calling="ArtifactRegistry")
        self.client = ClientHandler(
            self.artifactregistry_v1.ArtifactRegistryClient
        ).get()
        self.types = self.artifactregistry_v1.types
        self.FailedPrecondition = ModuleHandler(
            "google.api_core.exceptions"
        ).please_import("FailedPrecondition", who_is_calling="ArtifactRegistry")
        self.NotFound = ModuleHandler("google.api_core.exceptions").please_import(
            "NotFound", who_is_calling="ArtifactRegistry"
        )
        self.parent = self._get_parent()
        if self.project is None:
            raise ValueError("Project is required.")

    def __repr__(self):
        return f"ArtifactRegistry({self.project}/{self.location}/{self.path})"

    def _get_level(self, repository=None, image=None, version=None, tag=None):
        """
        Get the level of the path. Can be either 'project', 'location', 'repository', or 'file'.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image.
        - tag (str): The tag of the image.

        Returns:
        - str: The level of the path.
        """
        if tag or self.tag:
            return "tag"
        elif version or self.version:
            return "version"
        elif image or self.image:
            return "image"
        elif repository or self.repository:
            return "repository"
        elif self.location:
            return "location"
        elif self.project:
            return "project"
        else:
            return None

    def _get_path(self, shorten=True):
        """
        Get the path to the resource.

        Args:
        - shorten (bool): Whether to shorten the path to `project/location/repository/file`.

        Returns:
        - str: The path to the resource.
        """
        path = ""
        if self.level is None:
            return path
        elif self.level == "project":
            path = f"projects/{self.project}"
        elif self.level == "location":
            path = f"projects/{self.project}/locations/{self.location}"
        elif self.level == "repository":
            path = f"projects/{self.project}/locations/{self.location}/repositories/{self.repository}"
        elif self.level == "image":
            path = f"projects/{self.project}/locations/{self.location}/repositories/{self.repository}/packages/{self.image}"
        elif self.level == "tag":
            path = f"projects/{self.project}/locations/{self.location}/repositories/{self.repository}/packages/{self.image}/tags/{self.tag}"
        elif self.level == "version":
            path = f"projects/{self.project}/locations/{self.location}/repositories/{self.repository}/packages/{self.image}/versions/sha256:{self.version}"
        if shorten:
            path = "/".join(path.split("/")[1::2])
        return path

    def _get_parent(self):
        """
        Get the parent resource. This is used for making requests to the API. For us this is either location or project.

        Returns:
        - str: The parent resource.
        """
        if self.location is not None:
            return f"projects/{self.project}/locations/{self.location}"
        elif self.project is not None:
            return f"projects/{self.project}"
        else:
            return ""

    def ls_repositories(self, full_id=False):
        """
        List repositories in the Artifact Registry for a given location.

        Args:
        - full_id (bool): Whether to return the full repository ID or just the name.

        Returns:
        - list: The list of repositories.
        """
        output = self.client.list_repositories(parent=self.parent)
        output = [repository.name for repository in output]
        if not full_id:
            output = [repository.split("/")[-1] for repository in output]
        return output

    def ls_files(self, repository=None):
        """
        List files in a repository.

        Args:
        - repository (str): The name of the repository.

        Returns:
        - list: The list of files.
        """
        repository = repository or self.repository
        parent = f"{self.parent}/repositories/{repository}"
        output = self.client.list_files(parent=parent)
        output = [file.name for file in output]
        return output

    def ls_images(self, repository=None, image=None, full_id=False):
        """
        List images in a repository.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - full_id (bool): Whether to return the full image ID or just the name.

        Returns:
        - list: The list of images.
        """
        repository = repository or self.repository
        image = image or self.image
        parent = f"{self.parent}/repositories/{repository}"
        output = self.client.list_packages(parent=parent)
        output = [package.name for package in output]
        if not full_id:
            # Careful not to split("/") in case there is a "/" in the package name
            images = [package.replace(f"{parent}/packages/", "") for package in output]
            output = [f"{repository}/{image}" for image in images]
        return output

    def ls_versions(self, repository=None, image=None, tag=None, full_id=False):
        """
        List versions in an image.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - tag (str): The tag of the image.
        - full_id (bool): Whether to return the full tag ID or just the name.

        Returns:
        - list: The list of tags, sorted by creation time (latest first)
        """
        repository = repository or self.repository
        image = image or self.image
        tag = tag or self.tag
        parent = f"{self.parent}/repositories/{repository}/packages/{image}"
        output = self.client.list_versions(parent=parent)
        output = [(tag.name, tag.create_time) for tag in output]
        output = sorted(output, key=lambda x: x[1], reverse=True)
        output = [tag[0] for tag in output]
        if not full_id:
            versions = [tag.replace(f"{parent}/versions/", "") for tag in output]
            output = [f"{repository}/{image}/{version}" for version in versions]
        return output

    def ls_tags(self, repository=None, image=None, version=None, full_id=False):
        """
        List tags in a version.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image.
        - full_id (bool): Whether to return the full tag ID or just the name.

        Returns:
        - list: The list of tags.
        """
        repository = repository or self.repository
        image = image or self.image
        version = version or self.version
        parent = f"{self.parent}/repositories/{repository}/packages/{image}"
        output = self.client.list_tags(parent=parent)
        tag_version = f"{parent}/versions/sha256:{version}"
        output = [tag.name for tag in output if tag.version == tag_version]
        if not full_id:
            tags = [tag.replace(f"{parent}/tags/", "") for tag in output]
            output = [f"{repository}/{image}:{tag}" for tag in tags]
        return output

    def ls(self):
        """
        List repositories or files in a repository.

        Returns:
        - list: The list of repositories or files.
        """
        if self.level == "version":
            return self.ls_tags()
        elif self.level == "image":
            return self.ls_versions()
        elif self.level == "repository":
            return self.ls_images()
        elif self.level == "location":
            return self.ls_repositories()
        elif self.level == "project":
            raise ValueError("Cannot list items at the project level.")

    def get_repository(self, repository=None):
        """
        Get a repository.

        Args:
        - repository (str): The name of the repository.

        Returns:
        - Repository: The repository.
        """
        repository = repository or self.repository
        parent = f"{self.parent}/repositories/{repository}"
        return self.client.get_repository(name=parent)

    def get_image(self, repository=None, image=None):
        """
        Get an image.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.

        Returns:
        - Package: The image.
        """
        repository = repository or self.repository
        image = image or self.image
        parent = f"{self.parent}/repositories/{repository}/packages/{image}"
        return self.client.get_package(name=parent)

    def get_version(self, repository=None, image=None, version=None):
        """
        Get a version.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image.

        Returns:
        - Version: The version.
        """
        repository = repository or self.repository
        image = image or self.image
        version = version or self.version
        parent = f"{self.parent}/repositories/{repository}/packages/{image}/versions/sha256:{version}"
        return self.client.get_version(name=parent)

    def get_version_from_tag(self, repository=None, image=None, tag=None):
        """
        Get a tag.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - tag (str): The tag of the image.

        Returns:
        - Tag: The tag.
        """
        repository = repository or self.repository
        image = image or self.image
        tag = tag or self.tag
        parent = f"{self.parent}/repositories/{repository}/packages/{image}/tags/{tag}"
        result = self.client.get_tag(name=parent)
        if result is None:
            raise ValueError(f"Tag not found in {self.location}/{repository}/{image}.")
        version = result.version
        sha256_version = version.replace(
            f"{self.parent}/repositories/{repository}/packages/{image}/versions/sha256:",
            "",
        )
        return sha256_version

    def get_from_tag(self, repository=None, image=None, tag=None):
        """
        Get the artifact from a tag.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - tag (str): The tag of the image.

        Returns:
        - Tag: The tag.
        """
        version = self.get_version_from_tag(repository, image, tag)
        output = self.get_version(repository, image, version)
        return output

    def get_tag(self, repository=None, image=None, tag=None):
        """
        Get a tag.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - tag (str): The tag of the image.

        Returns:
        - Tag: The tag.
        """
        repository = repository or self.repository
        image = image or self.image
        tag = tag or self.tag
        parent = f"{self.parent}/repositories/{repository}/packages/{image}/tags/{tag}"
        return self.client.get_tag(name=parent)

    def get(self):
        """
        Get an image or version.

        Returns:
        - Package or Version: The image or version.
        """
        if self.level == "repository":
            return self.get_repository()
        elif self.level == "image":
            return self.get_image()
        elif self.level == "version":
            return self.get_version()
        elif self.level == "tag":
            return self.get_from_tag()
        else:
            raise ValueError("Cannot get item at this level.")

    def exists(self):
        """
        Check if a repo/image/version/tag exists.

        Returns:
        - bool: Whether the image or version exists.
        """
        try:
            self.get()
            return True
        except self.NotFound:
            return False

    def delete_repository(self, repository=None, **kwargs):
        """
        Delete a repository.

        Args:
        - repository (str): The name of the repository.

        Returns:
        - None
        """
        repository = repository or self.repository
        parent = f"{self.parent}/repositories/{repository}"
        output = self.client.delete_repository(name=parent)
        output.result()
        log(f"Artifact Registry - Deleted repository {self.location}/{repository}.")
        return output

    def delete_image(self, repository=None, image=None, **kwargs):
        """
        Delete an image.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.

        Returns:
        - None
        """
        repository = repository or self.repository
        image = image or self.image
        parent = f"{self.parent}/repositories/{repository}/packages/{image}"
        output = self.client.delete_package(name=parent)
        output.result()
        log(f"Artifact Registry - Deleted image {self.location}/{repository}/{image}.")
        return output

    def delete_version(self, repository=None, image=None, version=None, **kwargs):
        """
        Delete a version.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image.

        Returns:
        - None
        """
        repository = repository or self.repository
        image = image or self.image
        version = version or self.version
        parent = f"{self.parent}/repositories/{repository}/packages/{image}/versions/sha256:{version}"
        output = self.client.delete_version(name=parent)
        try:
            output.result()
        except self.FailedPrecondition as e:
            if "because it is tagged." in str(e):
                log("Artifact Registry - Version is tagged. Deleting tags first...")
                tags = self.ls_tags(repository, image, version)
                self._delete_parallel(tags)
                output = self.client.delete_version(name=parent)
        log(
            f"Artifact Registry - Deleted version {self.location}/{repository}/{image}/sha256:{version}"
        )
        return output

    def delete_tag(self, tag=None, repository=None, image=None, **kwargs):
        """
        Delete a tag.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - tag (str): The tag of the image.

        Returns:
        - None
        """
        repository = repository or self.repository
        image = image or self.image
        tag = tag or self.tag
        parent = f"{self.parent}/repositories/{repository}/packages/{image}/tags/{tag}"
        output = self.client.delete_tag(name=parent)
        try:
            output.result()
        except AttributeError:
            pass
        log(
            f"Artifact Registry - Deleted tag {self.location}/{repository}/{image}:{tag}."
        )
        return output

    def delete(
        self, repository=None, image=None, version=None, tag=None, errors="ignore"
    ):
        """
        Delete a repository, image, version, or a tag.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image.
        - tag (str): The tag of the image.
        - errors (str): How to handle errors if the item is not found. Can be either 'ignore' or 'raise'.

        Returns:
        - None
        """
        if self.level == "repository":
            operation = self.delete_repository
        elif self.level == "image":
            operation = self.delete_image
        elif self.level == "version":
            operation = self.delete_version
        elif self.level == "tag":
            operation = self.delete_tag
        else:
            raise ValueError("Cannot delete item at this level.")
        try:
            output = operation(
                repository=repository, image=image, version=version, tag=tag
            )
        except self.NotFound as e:
            if errors == "ignore":
                log(f"Artifact Registry - WARNING: {e}")
            else:
                raise e

    def _delete_parallel(self, items):
        """
        Deletes items in parallel using threading.

        Args:
        - items (list): List of items to delete.
        """
        threads = []
        for item in items:
            thread = threading.Thread(
                target=lambda item=item: ArtifactRegistry(
                    path=item, location=self.location
                ).delete()
            )
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()

    def create_repository(
        self,
        name=None,
        format="docker",
        mode="standard",
        immutable_tags=False,
        version_policy=None,
    ):
        """
        Create a repository.

        Args:
        - name (str): The name of the repository. If not provided, the repository name from the constructor will be used.
        - format (str): The format of the repository. Can be either 'docker' or 'maven' (default: 'docker')
        - mode (str): The mode of the repository. One of 'standard', 'remote' or 'virtual' (default: 'standard')
        - immutable_tags (bool): [Docker only] Whether to use immutable tags (default: False)
        - version_policy (str): [Maven only] The version policy for the repository - None, Snapshop or Release (default: None)



        Returns:
        - Repository: The repository.
        """
        maven_config = None
        docker_config = None
        repository = name or self.repository
        parent = f"{self.parent}/repositories/{repository}"
        if format == "docker":
            docker_config = self.types.Repository.DockerRepositoryConfig(
                immutable_tags=immutable_tags
            )
        elif format == "maven":
            if isinstance(version_policy, str):
                version_policy = version_policy.lower()
            version_policy = {None: 0, "none": 0, "snapshot": 1, "release": 2}[
                version_policy
            ]
            version_policy = self.types.Repository.MavenRepositoryConfig.VersionPolicy(
                version_policy
            )
            maven_config = self.types.Repository.MavenRepositoryConfig(
                version_policy=version_policy
            )
        mode = {"standard": 1, "remote": 2, "virtual": 3}.get(mode.lower(), None)
        if mode is None:
            raise ValueError("Mode must be either 'standard', 'remote' or 'virtual'.")
        format = {"docker": 1, "maven": 2}.get(format.lower(), None)
        if format is None:
            raise ValueError("Format must be either 'docker' or 'maven'.")
        format = self.types.Repository.Format(format)
        mode = self.types.Repository.Mode(mode)
        repository = self.types.Repository(
            name=parent,
            format_=format,
            mode=mode,
            docker_config=docker_config,
            maven_config=maven_config,
        )
        output = self.client.create_repository(
            parent=self.parent, repository=repository, repository_id=self.repository
        )
        try:
            output.result()
        except AttributeError:
            pass
        log(
            f"Artifact Registry - Created repository {self.project}/{self.location}/{repository}."
        )
        return output

    def create_tag(self, tag=None, repository=None, image=None, version=None):
        """
        Create a tag.

        Args:
        - tag (str): The tag to add to the version of the image.
        - value (str): The value of the tag.
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image.

        Returns:
        - Tag: The tag.
        """
        repository = repository or self.repository
        image = image or self.image
        version = version or self.version
        tag = tag or self.tag
        tag_name = tag
        Tag = self.types.Tag
        package_id = f"{self.parent}/repositories/{repository}/packages/{image}"
        version_id = f"{package_id}/versions/sha256:{version}"
        tag_id = f"{package_id}/tags/{tag}"
        tag = Tag(name=tag_id, version=version_id)
        output = self.client.create_tag(parent=package_id, tag=tag, tag_id=tag_name)
        log(
            f"Artifact Registry - Created tag {self.location}/{repository}/{image}:{tag}."
        )
        return output

    def create(self, repository=None, image=None, version=None, tag=None, **kwargs):
        """
        Create a repository, image, version, or a tag.

        Args:
        - repository (str): The name of the repository.
        - image (str): The name of the image.
        - version (str): The version of the image.
        - tag (str): The tag of the image.
        - kwargs: Additional arguments to pass to the creation function.

        Returns:
        - Repository, Package, Version, or Tag: The repository, image, version, or tag.
        """
        inputs = {
            "repository": repository,
            "image": image,
            "version": version,
            "tag": tag,
        }
        if self.level == "repository":
            return self.create_repository(**inputs, **kwargs)
        elif self.level == "image":
            raise ValueError("Creating images from the API is not supported.")
        elif self.level == "version":
            raise ValueError("Creating versions from the API is not supported.")
        elif self.level == "tag":
            return self.create_tag(**inputs, **kwargs)
        else:
            raise ValueError("Cannot create item at this level.")


if __name__ == "__main__":
    ArtifactRegistry(
        "gcr.io/example-service-123/sha256:411f06abcbda4b36a77c6e792e699b4eeb0193ebe441b6144f8fe42db6eada47",
        location="us",
    ).create_tag("latest_123")



PYPI_NAMES = {
    "gcsfs": "gcsfs",
    "google.auth": "google-auth",
    "google.oauth2": "google-auth-oauthlib",
    "google.cloud.core": "google-cloud-core",
    "google.cloud.run_v1": "google-cloud-run",
    "google.cloud.run_v2": "google-cloud-run",
    "google.cloud.storage": "google-cloud-storage",
    "google.cloud.logging": "google-cloud-logging",
    "google.cloud.pubsub_v1": "google-cloud-pubsub",
    "google.cloud.bigquery": "google-cloud-bigquery",
    "google.cloud.datastore": "google-cloud-datastore",
    "google.cloud.firestore": "google-cloud-firestore",
    "google.cloud.dataplex_v1": "google-cloud-dataplex",
    "google.cloud.functions_v1": "google-cloud-functions",
    "google.cloud.functions_v2": "google-cloud-functions",
    "google.cloud.scheduler_v1": "google-cloud-scheduler",
    "google.cloud.secretmanager": "google-cloud-secret-manager",
    "google.cloud.resource_manager": "google-cloud-resource-manager",
    "google.cloud.artifactregistry": "google-cloud-artifact-registry",
    "google.cloud.resourcemanager_v3": "google-cloud-resource-manager",
    "google.cloud.artifactregistry_v1": "google-cloud-artifact-registry",
    "yaml": "pyyaml",
    "pandas": "pandas",
    "docker": "docker",
    "pyarrow": "pyarrow",
    "pandas_gbq": "pandas-gbq",
}
DEFAULT_ARGS = {
    "location": "europe-west2",
}



from gcp_pal.utils import (
    log,
    ClientHandler,
    ModuleHandler,
    get_default_arg,
)


class Project:

    def __init__(self, project_id: str = None, folder: str = None):
        self.project_id = project_id or get_default_arg("project")
        self.folder = folder
        self.parent = f"folders/{self.folder}" if self.folder else None
        self.name = f"projects/{self.project_id}"

        self.resourcemanager = ModuleHandler("google.cloud").please_import(
            "resourcemanager_v3", who_is_calling="Project"
        )
        self.ProjectsClient = self.resourcemanager.ProjectsClient
        self.types = self.resourcemanager.types
        self.client = ClientHandler(self.ProjectsClient).get()

    def __repr__(self):
        return f"Project({self.project_id})"

    def create(self):
        """
        Creates a project.

        Returns:
        - google.cloud.resourcemanager_v3.types.Project
        """
        project = self.types.Project(project_id=self.project_id, parent=self.parent)
        output = self.client.create_project(project=project)
        output = output.result(timeout=300)
        log(f"Project - Created project '{self.project_id}'.")
        return output

    def delete(self):
        """
        Deletes a project. It will be marked for deletion, and will be deleted after 30 days.

        Returns:
        - None
        """
        self.client.delete_project(name=self.name)
        log(f"Project - Deleted project '{self.project_id}'.")

    def undelete(self):
        """
        Restores a project which has been marked for deletion.

        Returns:
        - None
        """
        self.client.undelete_project(name=self.name)
        log(f"Project - Restored project '{self.project_id}'.")

    def ls(self, active_only: bool = True):
        """
        Lists all projects which are available to the caller.

        Args:
        - active_only: (bool) If True, only return active projects. This filters out projects that are marked for deletion.

        Returns:
        - List of project IDs
        """
        projects = self.client.search_projects()
        projects = [x for x in projects]
        if active_only:
            projects = [x for x in projects if x.state.name == "ACTIVE"]
        project_ids = [x.project_id for x in projects]
        return project_ids

    def get(self):
        """
        Retrieves the project identified by the specified project ID.

        Returns:
        - google.cloud.resourcemanager_v3.types.Project
        """
        project = self.client.get_project(name=self.name)
        return project

    def number(self):
        """
        Retrieves the project number.

        Returns:
        - (str) Project number (e.g. '123456789012')
        """
        got = self.get()
        output = got.name.split("/")[-1]
        return output


if __name__ == "__main__":
    # print(Project().ls(active_only=True))
    print(Project("vitaminb16").get())
